/*
 * Copyright (c) 2003 Sun Microsystems, Inc. All Rights Reserved.
 * Copyright (c) 2010 JogAmp Community. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN
 * MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR
 * ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR
 * DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE
 * DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,
 * ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF
 * SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for use
 * in the design, construction, operation or maintenance of any nuclear
 * facility.
 *
 * Sun gratefully acknowledges that this software was originally authored
 * and developed by Kenneth Bradley Russell and Christopher John Kline.
 */

package com.jogamp.gluegen.cgram.types;

import com.jogamp.common.os.MachineDataInfo;
import com.jogamp.gluegen.GlueGen;
import com.jogamp.gluegen.TypeConfig;
import com.jogamp.gluegen.cgram.types.TypeComparator.SemanticEqualityOp;

/** Models a C type. Primitive types include int, float, and
    double. All types have an associated name. Structs and unions are
    modeled as "compound" types -- composed of fields of primitive or
    other types. */
public abstract class Type implements Cloneable, SemanticEqualityOp {
  public final boolean relaxedEqSem;
  private final int cvAttributes;
  private String name;
  private SizeThunk size;
  private int typedefedCVAttributes;
  private boolean hasTypedefName;
  private boolean hasCachedHash;
  private int cachedHash;
  private boolean hasCachedSemanticHash;
  private int cachedSemanticHash;

  protected Type(final String name, final SizeThunk size, final int cvAttributes) {
    setName(name);
    this.relaxedEqSem = TypeConfig.relaxedEqualSemanticsTest();
    this.cvAttributes = cvAttributes;
    this.size = size;
    this.typedefedCVAttributes = 0;
    this.hasTypedefName = false;
    this.hasCachedHash = false;
    this.cachedHash = 0;
    this.hasCachedSemanticHash = false;
    this.cachedSemanticHash = 0;
  }

  protected final void clearCache() {
    cachedHash = 0;
    hasCachedHash = false;
    cachedSemanticHash = 0;
    hasCachedHash = false;
  }

  @Override
  public Object clone() {
    try {
        return super.clone();
    } catch (final CloneNotSupportedException ex) {
        throw new InternalError();
    }
  }

  public final boolean isAnonymous() { return null == name; }

  public boolean hasName() { return null != name; }

  /** Returns the name of this type. The returned string is suitable
      for use as a type specifier for native C. Does not include any const/volatile
      attributes. */
  public final String getCName() { return getCName(false); }

  /** Returns the name of this type, optionally including
      const/volatile attributes. The returned string is suitable for
      use as a type specifier for native C. */
  public String getCName(final boolean includeCVAttrs) { return getName(includeCVAttrs); }

  /** Returns the name of this type. The returned string is suitable
      for use as a type specifier for Java. Does not include any const/volatile
      attributes. */
  public final String getName() { return getName(false); }

  /** Returns the name of this type, optionally including
      const/volatile attributes. The returned string is suitable for
      use as a type specifier for Java. */
  public String getName(final boolean includeCVAttrs) {
    if (!includeCVAttrs) {
      return name;
    }
    return getCVAttributesString() + name;
  }

  /**
   * Returns a string representation of this type.
   * The returned string is suitable for use as a type specifier for native C.
   * It does contain an expanded description of structs/unions,
   * hence may not be suitable for type declarations.
   */
  @Override
  public String toString() {
    return getCName(true);
  }


  private void append(final StringBuilder sb, final String val, final boolean prepComma) {
      if( prepComma ) {
          sb.append(", ");
      }
      sb.append(val);
  }
  // For debugging
  public String getDebugString() {
    final StringBuilder sb = new StringBuilder();
    boolean prepComma = false;
    sb.append("CType[");
    sb.append("(").append(getClass().getSimpleName()).append(") ");
    if( hasTypedefName() ) {
        sb.append("typedef ");
    }
    if( null != name ) {
        sb.append("'").append(name).append("'");
    } else {
        sb.append("ANON");
    }
    final Type targetType = getTargetType();
    if( null != targetType && this != targetType ) {
        sb.append(" -> ");
        if (!targetType.isFunction()) {
            sb.append(targetType.toString() + " * " + getCVAttributesString());
        } else {
            sb.append(((FunctionType) targetType).toString(null /* functionName */, null /* callingConvention */, false, true));
        }
    }
    if( GlueGen.debug() ) {
        // sb.append(", o=0x"+Integer.toHexString(objHash())+" h=0x"+Integer.toHexString(hashCode()));
        sb.append(", o=0x"+Integer.toHexString(objHash()));
    }
    sb.append(", size ");
    prepComma=true;
    if( null != size ) {
        final long mdSize;
        {
            long _mdSize = -1;
            try {
                _mdSize = size.computeSize(MachineDataInfo.StaticConfig.LP64_UNIX.md);
            } catch (final Exception e) {}
            mdSize = _mdSize;
        }
        sb.append("[fixed ").append(size.hasFixedNativeSize()).append(", lnx64 ").append(mdSize).append("]");
    } else {
        sb.append(" ZERO");
    }
    append(sb, "[", prepComma); prepComma=false;
    if( isConst() ) {
        append(sb, "const ", false);
    }
    if( isVolatile() ) {
        append(sb, "volatile ", false);
    }
    if( isPointer() ) {
        append(sb, "pointer*"+pointerDepth(), prepComma); prepComma=true;
    }
    if( isArray() ) {
        append(sb, "array*"+arrayDimension(), prepComma); prepComma=true;
    }
    if( isBit() ) {
        append(sb, "bit", prepComma); prepComma=true;
    }
    if( isCompound() ) {
        sb.append("struct{").append(asCompound().getStructName()).append(": ").append(asCompound().getNumFields());
        append(sb, "}", prepComma); prepComma=true;
    }
    if( isDouble() ) {
        append(sb, "double", prepComma); prepComma=true;
    }
    if( isEnum() ) {
        final EnumType eT = asEnum();
        sb.append("enum ").append(" [").append(eT.getUnderlyingType()).append("] {").append(eT.getNumEnumerates()).append(": ");
        eT.appendEnums(sb, false);
        prepComma=true;
    }
    if( isFloat() ) {
        append(sb, "float", prepComma); prepComma=true;
    }
    if( isFunction() ) {
        append(sb, "function", prepComma); prepComma=true;
    }
    if( isFunctionPointer() ) {
        append(sb, "funcPointer", prepComma); prepComma=true;
    }
    if( isInt() ) {
        append(sb, "int", prepComma); prepComma=true;
    }
    if( isVoid() ) {
        append(sb, "void", prepComma); prepComma=true;
    }
    sb.append("]]");
    return sb.toString();
  }
  private final int objHash() { return super.hashCode(); }


  protected final void setName(final String name) {
    if (name == null) {
      this.name = name;
    } else {
      this.name = name.intern();
    }
    clearCache();
  }

  /** Set the name of this type; used for handling typedefs. */
  public void setTypedefName(final String name) {
    setName(name);
    // Capture the const/volatile attributes at the time of typedef so
    // we don't redundantly repeat them in the CV attributes string
    typedefedCVAttributes = cvAttributes;
    hasTypedefName = true;
  }

  /** Indicates whether {@link #setTypedefName(String)} has been called on this type,
      indicating that it already has a typedef name. */
  public final boolean hasTypedefName() {
    return hasTypedefName;
  }

  /** SizeThunk which computes size of this type in bytes. */
  public SizeThunk    getSize()    { return size; }
  /** Size of this type in bytes according to the given MachineDataInfo. */
  public long         getSize(final MachineDataInfo machDesc) {
    final SizeThunk thunk = getSize();
    if (thunk == null) {
      throw new RuntimeException("No size set for type \"" + getName() + "\"");
    }
    return thunk.computeSize(machDesc);
  }
  /** Set the size of this type; only available for CompoundTypes. */
  void setSize(final SizeThunk size) {
      this.size = size;
      clearCache();
  }

  /** Casts this to a BitType or returns null if not a BitType. */
  public BitType      asBit()      { return null; }
  /** Casts this to an IntType or returns null if not an IntType. */
  public IntType      asInt()      { return null; }
  /** Casts this to an EnumType or returns null if not an EnumType. */
  public EnumType     asEnum()     { return null; }
  /** Casts this to a FloatType or returns null if not a FloatType. */
  public FloatType    asFloat()    { return null; }
  /** Casts this to a DoubleType or returns null if not a DoubleType. */
  public DoubleType   asDouble()   { return null; }
  /** Casts this to a PointerType or returns null if not a PointerType. */
  public PointerType  asPointer()  { return null; }
  /** Casts this to an ArrayType or returns null if not an ArrayType. */
  public ArrayType    asArray()    { return null; }
  /** Casts this to a CompoundType or returns null if not a CompoundType. */
  public CompoundType asCompound() { return null; }
  /** Casts this to a FunctionType or returns null if not a FunctionType. */
  public FunctionType asFunction() { return null; }
  /** Casts this to a VoidType or returns null if not a VoidType. */
  public VoidType     asVoid()     { return null; }

  /** Indicates whether this is a BitType. */
  public boolean      isBit()      { return (asBit()      != null); }
  /** Indicates whether this is an IntType. */
  public boolean      isInt()      { return (asInt()      != null); }
  /** Indicates whether this is an EnumType. */
  public boolean      isEnum()     { return (asEnum()     != null); }
  /** Indicates whether this is a FloatType. */
  public boolean      isFloat()    { return (asFloat()    != null); }
  /** Indicates whether this is a DoubleType. */
  public boolean      isDouble()   { return (asDouble()   != null); }
  /** Indicates whether this is a PointerType. */
  public boolean      isPointer()  { return (asPointer()  != null); }
  /** Indicates whether this is an ArrayType. */
  public boolean      isArray()    { return (asArray()    != null); }
  /** Indicates whether this is a CompoundType. */
  public boolean      isCompound() { return (asCompound() != null); }
  /** Indicates whether this is a FunctionType. */
  public boolean      isFunction() { return (asFunction() != null); }
  /** Indicates whether this is a VoidType. */
  public boolean      isVoid()     { return (asVoid()     != null); }

  /** Indicates whether this type is const. */
  public boolean      isConst()    { return (((cvAttributes & ~typedefedCVAttributes) & CVAttributes.CONST) != 0); }
  /** Indicates whether this type is volatile. */
  public boolean      isVolatile() { return (((cvAttributes & ~typedefedCVAttributes) & CVAttributes.VOLATILE) != 0); }

  /** Indicates whether this type is a primitive type. */
  public boolean      isPrimitive(){ return false; }

  /** Convenience routine indicating whether this Type is a pointer to
      a function. */
  public boolean isFunctionPointer() {
    return (isPointer() && asPointer().getTargetType().isFunction());
  }

  /** Hashcode for Types. */
  @Override
  public final int hashCode() {
    if( !hasCachedHash ) {
        // 31 * x == (x << 5) - x
        int hash = 31 + ( hasTypedefName ? 1 : 0 );
        hash = ((hash << 5) - hash) + ( null != size ? size.hashCode() : 0 );
        hash = ((hash << 5) - hash) + cvAttributes;
        hash = ((hash << 5) - hash) + ( null != name ? name.hashCode() : 0 );
        if( !hasTypedefName ) {
            hash = ((hash << 5) - hash) + hashCodeImpl();
        }
        cachedHash = hash;
        hasCachedHash = true;
    }
    return cachedHash;
  }
  protected abstract int hashCodeImpl();

  /**
   * Equality test for Types inclusive its given {@link #getName() name}.
   */
  @Override
  public final boolean equals(final Object arg) {
    if (arg == this) {
        return true;
    } else  if ( !getClass().isInstance(arg) ) { // implies null == arg || !(arg instanceof Type)
        return false;
    } else {
        final Type t = (Type)arg;
        if( hasTypedefName == t.hasTypedefName &&
            ( ( null != size && size.equals(t.size) ) ||
              ( null == size && null == t.size )
            ) &&
            cvAttributes == t.cvAttributes &&
            ( null == name ? null == t.name : name.equals(t.name) )
          )
        {
            if( !hasTypedefName ) {
                return equalsImpl(t);
            } else {
                return true;
            }
        } else {
            return false;
        }
    }
  }
  protected abstract boolean equalsImpl(final Type t);

  @Override
  public final int hashCodeSemantics() {
    if( !hasCachedSemanticHash ) {
        // 31 * x == (x << 5) - x
        int hash = 31 + ( null != size ? size.hashCodeSemantics() : 0 );
        if( !relaxedEqSem ) {
            hash = ((hash << 5) - hash) + cvAttributes;
        }
        hash = ((hash << 5) - hash) + hashCodeSemanticsImpl();
        cachedSemanticHash = hash;
        hasCachedSemanticHash = true;
    }
    return cachedSemanticHash;
  }
  protected abstract int hashCodeSemanticsImpl();

  @Override
  public final boolean equalSemantics(final SemanticEqualityOp arg) {
    if (arg == this) {
        return true;
    } else  if ( !(arg instanceof Type) ||
                 !getClass().isInstance(arg) ) { // implies null == arg
        return false;
    } else {
        final Type t = (Type) arg;
        if( ( ( null != size && size.equalSemantics(t.size) ) ||
              ( null == size && null == t.size )
            ) &&
            ( relaxedEqSem || cvAttributes == t.cvAttributes )
          )
        {
            return equalSemanticsImpl(t);
        } else {
            return false;
        }
    }
  }
  protected abstract boolean equalSemanticsImpl(final Type t);

  /** Visit this type and all of the component types of this one; for
      example, the return type and argument types of a FunctionType. */
  public void visit(final TypeVisitor visitor) {
    visitor.visitType(this);
  }

  public final int getCVAttributes() {
    return cvAttributes;
  }

  /** Returns a string indicating the const/volatile attributes of
      this type. */
  public final String getCVAttributesString() {
    if (isConst() && isVolatile()) return "const volatile ";
    if (isConst()) return "const ";
    if (isVolatile()) return "volatile ";
    return "";
  }

  /** Return a variant of this type matching the given const/volatile
      attributes. May return this object if the attributes match. */
  public final Type getCVVariant(final int cvAttributes) {
    if (this.cvAttributes == cvAttributes) {
      return this;
    }
    return newCVVariant(cvAttributes);
  }

  /** Create a new variant of this type matching the given
      const/volatile attributes. */
  abstract Type newCVVariant(int cvAttributes);

  /** Helper method for determining how many pointer indirections this
      type represents (i.e., "void **" returns 2). Returns 0 if this
      type is not a pointer type. */
  public int pointerDepth() {
    final PointerType pt = asPointer();
    if (pt == null) {
      return 0;
    }
    return 1 + pt.getTargetType().pointerDepth();
  }

  /** Helper method for determining how many array dimentions this
      type represents (i.e., "char[][]" returns 2). Returns 0 if this
      type is not an array type. */
  public int arrayDimension() {
    final ArrayType arrayType = asArray();
    if (arrayType == null) {
      return 0;
    }
    return 1 + arrayType.getElementType().arrayDimension();
  }

  /**
   * Helper method to returns the bottom-most element type of this type.
   * <p>
   * If this is a multidimensional array or pointer method returns the bottom-most element type,
   * otherwise this.
   * </p>
   */
  public Type getBaseElementType() {
      return this;
  }

  /**
   * Helper method to returns the target type of this type, in case another type is being referenced.
   */
  public Type getTargetType() {
      return this;
  }
}
